\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Core.Everything

open import Cubical.Foundations.Everything

open import Cubical.Data.Nat
open import Cubical.Data.Bool
open import Cubical.Data.Empty as ⊥

open import Cubical.Relation.Nullary
\end{code}

%<*Pos>
\begin{code}
data Pos : Set where
  pos1  : Pos
  x0    : Pos → Pos
  x1    : Pos → Pos
\end{code}
%</Pos>

\begin{code}
sucPos : Pos → Pos
sucPos pos1     = x0 pos1
sucPos (x0 ps)  = x1 ps
sucPos (x1 ps)  = x0 (sucPos ps)

-- Conversion of Pos to ℕ.

Pos→ℕ : Pos → ℕ
Pos→ℕ pos1     = suc zero
Pos→ℕ (x0 ps)  = doubleℕ (Pos→ℕ ps)
Pos→ℕ (x1 ps)  = suc (doubleℕ (Pos→ℕ ps))

-- Unary number induction on Pos.

posInd : {P : Pos → Set} → P pos1 → ((p : Pos) → P p → P (sucPos p)) → (p : Pos) → P p
posInd {P} h1 hs = f
  where
  H : (p : Pos) → P (x0 p) → P (x0 (sucPos p))
  H p hx0p  = hs (x1 p) (hs (x0 p) hx0p)

  f : (ps : Pos) → P ps
  f pos1     = h1
  f (x0 ps)  = posInd (hs pos1 h1) H ps
  f (x1 ps)  = hs (x0 ps) (posInd (hs pos1 h1) H ps)

-- Pos→ℕ is suc-morphism.

Pos→ℕsucPos : (p : Pos) → Pos→ℕ (sucPos p) ≡ suc (Pos→ℕ p)
Pos→ℕsucPos pos1    = refl
Pos→ℕsucPos (x0 p)  = refl
Pos→ℕsucPos (x1 p)  = λ i → doubleℕ (Pos→ℕsucPos p i)

-- zero is not in the image of Pos→ℕ.

zero≠Pos→ℕ : (p : Pos) → ¬ (zero ≡ Pos→ℕ p)
zero≠Pos→ℕ p  = posInd (λ prf → znots prf) hs p
  where
  hs : (p : Pos) → ¬ (zero ≡ Pos→ℕ p) → zero ≡ Pos→ℕ (sucPos p) → ⊥
  hs p neq ieq  = ⊥.rec (znots (ieq ∙ Pos→ℕsucPos p))

-- Conversion from ℕ to Pos.

ℕ→Pos : ℕ → Pos
ℕ→Pos zero           = pos1
ℕ→Pos (suc zero)     = pos1
ℕ→Pos (suc (suc n))  = sucPos (ℕ→Pos (suc n))

-- ℕ→Pos is a suc-morphism on { n | n ≥ 1 }.

ℕ→PosSuc : ∀ n → ¬ (zero ≡ n) → ℕ→Pos (suc n) ≡ sucPos (ℕ→Pos n)
ℕ→PosSuc zero neq     = ⊥.elim (neq refl)
ℕ→PosSuc (suc n) neq  = refl

-- Isomorphism, proven by unary number induction.

Pos→ℕ→Pos : (p : Pos) → ℕ→Pos (Pos→ℕ p) ≡ p
Pos→ℕ→Pos p  = posInd refl hs p
  where
  hs : (p : Pos) → ℕ→Pos (Pos→ℕ p) ≡ p → ℕ→Pos (Pos→ℕ (sucPos p)) ≡ sucPos p
  hs p hp  =
    ℕ→Pos (Pos→ℕ (sucPos p)) ≡⟨ cong ℕ→Pos (Pos→ℕsucPos p) ⟩
    ℕ→Pos (suc (Pos→ℕ p))    ≡⟨ ℕ→PosSuc (Pos→ℕ p) (zero≠Pos→ℕ p) ⟩
    sucPos (ℕ→Pos (Pos→ℕ p)) ≡⟨ cong sucPos hp ⟩
    sucPos p ∎

ℕ→Pos→ℕ : (n : ℕ) → Pos→ℕ (ℕ→Pos (suc n)) ≡ suc n
ℕ→Pos→ℕ zero     = refl
ℕ→Pos→ℕ (suc n)  =
  Pos→ℕ (sucPos (ℕ→Pos (suc n))) ≡⟨ Pos→ℕsucPos (ℕ→Pos (suc n)) ⟩
  suc (Pos→ℕ (ℕ→Pos (suc n)))    ≡⟨ cong suc (ℕ→Pos→ℕ n) ⟩
  suc (suc n) ∎

-- Binary numbers
\end{code}

%<*Bin>
\begin{code}
data Bin : Set where
  bin0    : Bin
  binPos  : Pos → Bin
\end{code}
%</Bin>

%<*six>
\begin{code}
six : Bin
six = binPos (x0 (x1 pos1))
\end{code}
%</six>

\begin{code}
ℕ→Bin : ℕ → Bin
ℕ→Bin zero     = bin0
ℕ→Bin (suc n)  = binPos (ℕ→Pos (suc n))

Bin→ℕ : Bin → ℕ
Bin→ℕ bin0        = zero
Bin→ℕ (binPos x)  = Pos→ℕ x

ℕ→Bin→ℕ : (n : ℕ) → Bin→ℕ (ℕ→Bin n) ≡ n
ℕ→Bin→ℕ zero           = refl
ℕ→Bin→ℕ (suc zero)     = refl
ℕ→Bin→ℕ (suc (suc n))  =
    Pos→ℕ (sucPos (ℕ→Pos (suc n))) ≡⟨ Pos→ℕsucPos (ℕ→Pos (suc n)) ⟩
    suc (Pos→ℕ (ℕ→Pos (suc n)))    ≡⟨ cong suc (ℕ→Bin→ℕ (suc n)) ⟩
    suc (suc n) ∎

Bin→ℕ→Bin : (n : Bin) → ℕ→Bin (Bin→ℕ n) ≡ n
Bin→ℕ→Bin bin0  = refl
Bin→ℕ→Bin (binPos p)  = posInd refl (λ p _ → rem p) p
  where
  rem : (p : Pos) → ℕ→Bin (Pos→ℕ (sucPos p)) ≡ binPos (sucPos p)
  rem p  =
    ℕ→Bin (Pos→ℕ (sucPos p))       ≡⟨ cong ℕ→Bin (Pos→ℕsucPos p) ⟩
    binPos (ℕ→Pos (suc (Pos→ℕ p))) ≡⟨ cong binPos (ℕ→PosSuc (Pos→ℕ p) (zero≠Pos→ℕ p) ∙
                                                              (cong sucPos (Pos→ℕ→Pos p))) ⟩
    binPos (sucPos p) ∎

sixℕ = Bin→ℕ six
\end{code}

%<*BinEqN>
\begin{code}
ℕ≃Bin : ℕ ≃ Bin
ℕ≃Bin  = isoToEquiv (iso ℕ→Bin Bin→ℕ Bin→ℕ→Bin ℕ→Bin→ℕ)

ℕ≡Bin : ℕ ≡ Bin
ℕ≡Bin  = ua ℕ≃Bin
\end{code}
%</BinEqN>

\begin{code}
sucBin : Bin → Bin
sucBin bin0        = binPos pos1
sucBin (binPos x)  = binPos (sucPos x)

Bin→ℕsuc : (x : Bin) → suc (Bin→ℕ x) ≡ Bin→ℕ (sucBin x)
Bin→ℕsuc bin0        = refl
Bin→ℕsuc (binPos x)  = sym (Pos→ℕsucPos x)

doubleBin : Bin → Bin
doubleBin bin0        = bin0
doubleBin (binPos x)  = binPos (x0 x)

Bin→ℕdouble : (x : Bin) → doubleℕ (Bin→ℕ x) ≡ Bin→ℕ (doubleBin x)
Bin→ℕdouble bin0        = refl
Bin→ℕdouble (binPos x)  = refl

Bin→ℕ-inj : (x y : Bin) → Bin→ℕ x ≡ Bin→ℕ y → x ≡ y
Bin→ℕ-inj x y p = sym (Bin→ℕ→Bin x) ∙∙ cong ℕ→Bin p ∙∙ Bin→ℕ→Bin y

ℕ→Bindouble : (x : ℕ) → doubleBin (ℕ→Bin x) ≡ ℕ→Bin (doubleℕ x)
ℕ→Bindouble x = Bin→ℕ-inj (doubleBin (ℕ→Bin x)) (ℕ→Bin (doubleℕ x)) (sym (Bin→ℕdouble (ℕ→Bin x))
              ∙ cong doubleℕ (ℕ→Bin→ℕ x) ∙ sym (ℕ→Bin→ℕ (doubleℕ x)))

-- TODO: upstream
doubleℕ-+ : (x y : ℕ) → doubleℕ (x + y) ≡ doubleℕ x + doubleℕ y
doubleℕ-+ zero y = refl
doubleℕ-+ (suc x) y = cong (suc ∘ suc) (doubleℕ-+ x y)

+-1 : (n : ℕ) → n + 1 ≡ suc n
+-1 n = +-suc n 0 ∙ cong suc (+-zero n)

\end{code}

%<*Binadd>
\begin{code}
_+Bin_ : Bin → Bin → Bin
_+Bin_  = transport (λ i → ℕ≡Bin i → ℕ≡Bin i → ℕ≡Bin i) _+_
\end{code}
%</Binadd>

%<*Binaddex>
\begin{code}
_ : (binPos (x0 (x0 pos1))) +Bin (binPos pos1) ≡ binPos (x1 (x0 pos1))
_  = refl
\end{code}
%</Binaddex>

-- Transport addition to binary
%<*Binaddp>
\begin{code}
addp    :  PathP  (λ i → ℕ≡Bin i → ℕ≡Bin i → ℕ≡Bin i) _+_ _+Bin_
addp i  =  transp  (λ j → ℕ≡Bin (i ∧ j) → ℕ≡Bin (i ∧ j) → ℕ≡Bin (i ∧ j)) (~ i) _+_

+Bin-assoc : (m n o : Bin) → m +Bin (n +Bin o) ≡ (m +Bin n) +Bin o
+Bin-assoc =
  transport  (λ i → (m n o : ℕ≡Bin i) →  addp i m (addp i n o)
                                         ≡ addp i (addp i m n) o)
             +-assoc
\end{code}
%</Binaddp>

%<*Binassoc>
\begin{code}
T : Set → Set
T X = Σ[ _+_ ∈ (X → X → X) ] ((x y z : X) → x + (y + z) ≡ (x + y) + z)

TBin : T Bin
TBin = subst T ℕ≡Bin (_+_ , +-assoc)

_+Bin'_ : Bin → Bin → Bin
_+Bin'_  = fst TBin

+Bin'-assoc : (m n o : Bin) → m +Bin' (n +Bin' o) ≡ (m +Bin' n) +Bin' o
+Bin'-assoc = snd TBin
\end{code}
%</Binassoc>
-- Efficient addition of binary numbers

%<*Binaddfastc1>
\begin{code}
mutual
  _+P_ : Pos → Pos → Pos
  pos1  +P y     = sucPos y
  x0 x  +P pos1  = x1 x
  x0 x  +P x0 y  = x0 (x +P y)
  x0 x  +P x1 y  = x1 (x +P y)
  x1 x  +P pos1  = x0 (sucPos x)
  x1 x  +P x0 y  = x1 (x +P y)
  x1 x  +P x1 y  = x0 (x +PC y)

  _+B_ : Bin → Bin → Bin
  bin0      +B y         = y
  x         +B bin0      = x
  binPos x  +B binPos y  = binPos (x +P y)
\end{code}
%</Binaddfastc1>

%<*Binaddfastc2>
\begin{code}
  -- Add with carry
  _+PC_ : Pos → Pos → Pos
  pos1  +PC pos1  = x1 pos1
  pos1  +PC x0 y  = x0 (sucPos y)
  pos1  +PC x1 y  = x1 (sucPos y)
  x0 x  +PC pos1  = x0 (sucPos x)
  x0 x  +PC x0 y  = x1 (x +P y)
  x0 x  +PC x1 y  = x0 (x +PC y)
  x1 x  +PC pos1  = x1 (sucPos x)
  x1 x  +PC x0 y  = x0 (x +PC y)
  x1 x  +PC x1 y  = x1 (x +PC y)
\end{code}
%</Binaddfastc2>

\begin{code}
-- Correctness:
+PC-suc : (x y : Pos) → x +PC y ≡ sucPos (x +P y)
+PC-suc pos1 pos1     = refl
+PC-suc pos1 (x0 y)   = refl
+PC-suc pos1 (x1 y)   = refl
+PC-suc (x0 x) pos1   = refl
+PC-suc (x0 x) (x0 y) = refl
+PC-suc (x0 x) (x1 y) = cong x0 (+PC-suc x y)
+PC-suc (x1 x) pos1   = refl
+PC-suc (x1 x) (x0 y) = cong x0 (+PC-suc x y)
+PC-suc (x1 x) (x1 y) = refl

sucPos-+P : (x y : Pos) → sucPos (x +P y) ≡ sucPos x +P y
sucPos-+P pos1 pos1     = refl
sucPos-+P pos1 (x0 y)   = refl
sucPos-+P pos1 (x1 y)   = refl
sucPos-+P (x0 x) pos1   = refl
sucPos-+P (x0 x) (x0 y) = refl
sucPos-+P (x0 x) (x1 y) = cong x0 (sym (+PC-suc x y))
sucPos-+P (x1 x) pos1   = refl
sucPos-+P (x1 x) (x0 y) = cong x0 (sucPos-+P x y)
sucPos-+P (x1 x) (x1 y) = cong x1 (+PC-suc  x y ∙ sucPos-+P x y)

ℕ→Pos-+P : (x y : ℕ) → ℕ→Pos (suc x + suc y) ≡ ℕ→Pos (suc x) +P ℕ→Pos (suc y)
ℕ→Pos-+P zero _    = refl
ℕ→Pos-+P (suc x) y = cong sucPos (ℕ→Pos-+P x y) ∙ sucPos-+P (ℕ→Pos (suc x)) (ℕ→Pos (suc y))
\end{code}

%<*Binaddfast2>
\begin{code}
ℕ→Bin-+B : (x y : ℕ) → ℕ→Bin (x + y) ≡ ℕ→Bin x +B ℕ→Bin y
\end{code}
%</Binaddfast2>
\begin{code}
ℕ→Bin-+B zero y          = refl
ℕ→Bin-+B (suc x) zero    = cong (binPos ∘ ℕ→Pos ∘ suc) (+-zero x)
ℕ→Bin-+B (suc x) (suc y) = cong binPos (ℕ→Pos-+P x y)

-- This amounts to proving that (x y : Bin) → x +B y ≡ ℕ→Bin (Bin→ℕ x + Bin→ℕ y)
\end{code}
%<*Binaddfast3>
\begin{code}
+B≡+Bin : _+B_ ≡ _+Bin_
+B≡+Bin i x y = goal x y i
  where
  goal : (x y : Bin) → x +B y ≡ ℕ→Bin (Bin→ℕ x + Bin→ℕ y)
  goal x y =  (λ i → Bin→ℕ→Bin x (~ i) +B Bin→ℕ→Bin y (~ i))
           ∙  sym (ℕ→Bin-+B (Bin→ℕ x) (Bin→ℕ y))
\end{code}
%</Binaddfast3>
\begin{code}

-- Alternative more complex proof by induction on Bin and Pos:

-- Pos→ℕ-+P : (x y : Pos) → Pos→ℕ (x +P y) ≡ Pos→ℕ x + Pos→ℕ y
-- Pos→ℕ-+P pos1 y        = Pos→ℕsucPos y
-- Pos→ℕ-+P (x0 x) pos1   = sym (+-1 _)
-- Pos→ℕ-+P (x0 x) (x0 y) = cong doubleℕ (Pos→ℕ-+P x y) ∙ doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y)
-- Pos→ℕ-+P (x0 x) (x1 y) = cong (suc ∘ doubleℕ) (Pos→ℕ-+P x y)
--                       ∙∙ cong suc (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))
--                       ∙∙ sym (+-suc _ _)
-- Pos→ℕ-+P (x1 x) pos1   = cong doubleℕ (Pos→ℕsucPos x) ∙ sym (+-1 _)
-- Pos→ℕ-+P (x1 x) (x0 y) = cong (suc ∘ doubleℕ) (Pos→ℕ-+P x y)
--                        ∙ cong suc (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))
-- Pos→ℕ-+P (x1 x) (x1 y) = cong (doubleℕ ∘ Pos→ℕ) (+PC-suc x y)
--                        ∙ cong doubleℕ (Pos→ℕsucPos (x +P y))
--                        ∙ cong (suc ∘ suc ∘ doubleℕ) (Pos→ℕ-+P x y)
--                        ∙ cong (suc ∘ suc) (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))
--                        ∙ cong suc (sym (+-suc _ _))

-- -- This amounts to proving that (x y : Bin) → x +B y ≡ ℕ→Bin (Bin→ℕ x + Bin→ℕ y)
-- +B≡+Bin : _+B_ ≡ _+Bin_
-- +B≡+Bin i bin0 y = Bin→ℕ→Bin y (~ i)
-- +B≡+Bin i (binPos x) bin0 = (sym (Bin→ℕ→Bin (binPos x)) ∙ sym (cong ℕ→Bin (+-zero (Pos→ℕ x)))) i
-- +B≡+Bin i (binPos x) (binPos y) = goal x y i
--   where
--   goal : (x y : Pos) → binPos (x +P y) ≡ ℕ→Bin (Pos→ℕ x + Pos→ℕ y)
--   goal pos1 y = sym (cong binPos (ℕ→PosSuc (Pos→ℕ y) (zero≠Pos→ℕ y) ∙ cong sucPos (Pos→ℕ→Pos y)))
--   goal (x0 x) pos1 = sym (Bin→ℕ→Bin (binPos (x1 x)))
--                    ∙ sym (cong ℕ→Bin (+-1 _))
--   goal (x0 x) (x0 y) = cong doubleBin (goal x y)
--                     ∙∙ ℕ→Bindouble _
--                     ∙∙ cong ℕ→Bin (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))
--   goal (x0 x) (x1 y) =
--     binPos (x0 x +P x1 y)
--       ≡⟨ cong binPos (sym (Pos→ℕ→Pos _)) ⟩
--     binPos (ℕ→Pos (Pos→ℕ (x1 (x +P y))))
--       ≡⟨ refl ⟩
--     binPos (ℕ→Pos (suc (doubleℕ (Pos→ℕ (x +P y)))))
--        ≡⟨ cong (binPos ∘ ℕ→Pos ∘ suc ∘ doubleℕ) (Pos→ℕ-+P x y) ⟩
--     binPos (ℕ→Pos (suc (doubleℕ (Pos→ℕ x + Pos→ℕ y))))
--       ≡⟨ sym (cong ℕ→Bin (+-suc _ _ ∙ cong suc (sym (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))))) ⟩
--     ℕ→Bin (Pos→ℕ (x0 x) + Pos→ℕ (x1 y)) ∎
--   goal (x1 x) pos1 =
--     cong binPos (sym (Pos→ℕ→Pos (x0 (sucPos x)))
--                ∙ cong (ℕ→Pos ∘ doubleℕ) (Pos→ℕsucPos x)
--                ∙ cong ℕ→Pos (sym (cong suc (+-suc (doubleℕ (Pos→ℕ x)) 0
--                             ∙ cong suc (+-zero _)))))
--   goal (x1 x) (x0 y) = sym (Bin→ℕ→Bin (binPos (x1 (x +P y))))
--                      ∙ cong (binPos ∘ ℕ→Pos ∘ suc)
--                             (cong doubleℕ (Pos→ℕ-+P x y)
--                              ∙ doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y))
--   goal (x1 x) (x1 y) =
--     binPos (x0 (x +PC y))
--       ≡⟨ cong (binPos ∘ x0) (+PC-suc x y) ⟩
--     binPos (x0 (sucPos (x +P y)))
--       ≡⟨ sym (Bin→ℕ→Bin (binPos (x0 (sucPos (x +P y))))) ⟩
--     ℕ→Bin (Bin→ℕ (binPos (x0 (sucPos (x +P y)))))
--       ≡⟨ cong (ℕ→Bin ∘ doubleℕ) (Pos→ℕsucPos (x +P y)) ⟩
--     ℕ→Bin (doubleℕ (suc (Pos→ℕ (x +P y))))
--       ≡⟨ refl ⟩
--     binPos (ℕ→Pos (suc (suc (doubleℕ (Pos→ℕ (x +P y))))))
--       ≡⟨ cong (binPos ∘ ℕ→Pos ∘ suc ∘ suc ∘ doubleℕ) (Pos→ℕ-+P x y) ⟩
--     binPos (ℕ→Pos (suc (suc (doubleℕ (Pos→ℕ x + Pos→ℕ y)))))
--       ≡⟨ cong (binPos ∘ ℕ→Pos ∘ suc ∘ suc) (doubleℕ-+ (Pos→ℕ x) (Pos→ℕ y)) ⟩
--     binPos (ℕ→Pos (suc (suc (doubleℕ (Pos→ℕ x) + doubleℕ (Pos→ℕ y)))))
--       ≡⟨ cong (binPos ∘ ℕ→Pos) (sym (+-suc (suc (doubleℕ (Pos→ℕ x))) _)) ⟩
--     ℕ→Bin (Pos→ℕ (x1 x) + Pos→ℕ (x1 y)) ∎

\end{code}

%<*Binadd2>
\begin{code}
+B-assoc : (m n o : Bin) → m +B (n +B o) ≡ (m +B n) +B o
+B-assoc m n o =  (λ i → +B≡+Bin i m (+B≡+Bin i n o))
               ∙  +Bin-assoc m n o
               ∙  (λ i → +B≡+Bin (~ i) (+B≡+Bin (~ i) m n) o)
\end{code}
%</Binadd2>


-- Doubling experiment: we define a notion of "doubling structure" and
-- transport a proof that is proved directly using refl for binary
-- numbers to unary numbers. This is an example of program/data
-- refinement: we can use univalence to prove properties about
-- inefficient data-structures using efficient ones.

%<*doublestruct>
\begin{code}
record Double (A : Set) : Set where
  constructor doubleStruct
  field
    double : A → A
    elt : A
\end{code}
%</doublestruct>

\begin{code}
open Double

bin1024 : Bin
bin1024  = binPos (x0 (x0 (x0 (x0 (x0 (x0 (x0 (x0 (x0 (x0 pos1))))))))))
\end{code}

%<*doubleinstN>
\begin{code}
Doubleℕ : Double ℕ
Doubleℕ  = doubleStruct doubleℕ 1024
\end{code}
%</doubleinstN>

%<*doubleinstBin>
\begin{code}
DoubleBin : Double Bin
DoubleBin  = doubleStruct doubleBin bin1024
\end{code}
%</doubleinstBin>

\begin{code}
-- As these function don't commute strictly we have to prove it
-- separately and insert it in the proof of DoubleBin≡Doubleℕ below
-- (just like we had to in NatImplℕ≡NatImplBin

Bin≡ℕ : Bin ≡ ℕ
Bin≡ℕ  = isoToPath (iso Bin→ℕ ℕ→Bin ℕ→Bin→ℕ Bin→ℕ→Bin)

\end{code}

%<*doubleeq>
\begin{code}
DoubleBin≡Doubleℕ : PathP (λ i → Double (Bin≡ℕ i)) DoubleBin Doubleℕ
\end{code}
%</doubleeq>
\begin{code}
double (DoubleBin≡Doubleℕ i)  =
  λ x → glue (λ { (i  = i0) → doubleBin x
                ; (i  = i1) → doubleℕ x })
             (hcomp (λ j → λ { (i  = i0) → Bin→ℕdouble x j
                             ; (i  = i1) → doubleℕ x })
                    (doubleℕ (unglue (i ∨ ~ i) x)))
elt (DoubleBin≡Doubleℕ i)  = transp (λ j → Bin≡ℕ (i ∨ ~ j)) i (elt Doubleℕ)
\end{code}
%<*propdouble>
\begin{code}
doubles : {A : Set} (D : Double A) → ℕ → A → A
doubles D n x  = iter n (double D) x

record propDouble {A : Set} (D : Double A) : Set where
  field
    proof : doubles D 20 (elt D) ≡ doubles D 5 (doubles D 15 (elt D))
\end{code}
%</propdouble>

\begin{code}
open propDouble
\end{code}

%<*propdoublen>
\begin{code}
propDoubleBin : propDouble DoubleBin
proof propDoubleBin  = refl

propDoubleℕ : propDouble Doubleℕ
propDoubleℕ = transport  (λ i → propDouble (DoubleBin≡Doubleℕ i))
                         propDoubleBin
\end{code}
%</propdoublen>
