\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Foundations.Prelude
open import Cubical.Foundations.Equiv
open import Cubical.Foundations.Structure
open import Cubical.Foundations.Isomorphism
open import Cubical.Foundations.Univalence

open import Cubical.Functions.FunExtEquiv
open import Cubical.Data.Nat hiding (_+_ ; +-comm)
import Cubical.Data.Empty as ⊥
open import Cubical.Relation.Nullary
open import Cubical.Algebra.CommRing

infixr 5 _∷_

private
  variable
    ℓ : Level
\end{code}

%<*vec>
\begin{code}
data Vec (A : Set ℓ) : ℕ → Set ℓ where
  []  : Vec A zero
  _∷_ : ∀ {n} (x : A) (xs : Vec A n) → Vec A (suc n)
\end{code}
%</vec>

%<*fin>
\begin{code}
data Fin : ℕ → Set where
  zero : {n : ℕ} → Fin (suc n)
  suc  : {n : ℕ} (i : Fin n) → Fin (suc n)
\end{code}
%</fin>

%<*matrix>
\begin{code}
FinMatrix : (A : Set ℓ) (m n : ℕ) → Set ℓ
FinMatrix A m n = Fin m → Fin n → A

VecMatrix : (A : Set ℓ) (m n : ℕ) → Set ℓ
VecMatrix A m n = Vec (Vec A n) m
\end{code}
%</matrix>

%<*matrix2>
\begin{code}
FinVec→Vec : {A : Set ℓ} {n : ℕ} → (Fin n → A) → Vec A n
FinVec→Vec {n = zero}  xs = []
FinVec→Vec {n = suc _} xs = xs zero ∷ FinVec→Vec (λ x → xs (suc x))

lookup : {A : Set ℓ} {n : ℕ} → Fin n → Vec A n → A
lookup zero    (x ∷ xs) = x
lookup (suc i) (x ∷ xs) = lookup i xs

Fin→VecMatrix : {A : Set ℓ} {m n : ℕ} → FinMatrix A m n → VecMatrix A m n
Fin→VecMatrix M = FinVec→Vec (λ fm → FinVec→Vec (λ fn → M fm fn))

Vec→FinMatrix : {A : Set ℓ} {m n : ℕ} → VecMatrix A m n → FinMatrix A m n
Vec→FinMatrix M fn fm = lookup fm (lookup fn M)
\end{code}
%</matrix2>

\begin{code}
Vec→FinVec : {A : Set ℓ} {n : ℕ} → Vec A n → (Fin n → A)
Vec→FinVec xs f = lookup f xs

¬Fin0 : ¬ Fin 0
¬Fin0 ()

Vec→FinVec→Vec : {A : Set ℓ} {n : ℕ} (xs : Vec A n) → FinVec→Vec (Vec→FinVec xs) ≡ xs
Vec→FinVec→Vec {n = zero}  [] = refl
Vec→FinVec→Vec {n = suc n} (x ∷ xs) i = x ∷ Vec→FinVec→Vec xs i

FinVec→Vec→FinVec : {A : Set ℓ} {n : ℕ} (xs : Fin n → A) → Vec→FinVec (FinVec→Vec xs) ≡ xs
FinVec→Vec→FinVec {n = zero} xs = funExt λ f → ⊥.rec (¬Fin0 f)
FinVec→Vec→FinVec {n = suc n} xs = funExt goal
  where
  goal : (f : Fin (suc n))
       → Vec→FinVec (xs zero ∷ FinVec→Vec (λ x → xs (suc x))) f ≡ xs f
  goal zero = refl
  goal (suc f) i = FinVec→Vec→FinVec (λ x → xs (suc x)) i f

FinMatrix→VecMatrix→FinMatrix : {A : Set ℓ} {m n : ℕ} (M : FinMatrix A m n) → Vec→FinMatrix (Fin→VecMatrix M) ≡ M
FinMatrix→VecMatrix→FinMatrix {m = zero} M = funExt λ f → ⊥.rec (¬Fin0 f)
FinMatrix→VecMatrix→FinMatrix {n = zero} M = funExt₂ λ _ f → ⊥.rec (¬Fin0 f)
FinMatrix→VecMatrix→FinMatrix {m = suc m} {n = suc n} M = funExt₂ goal
  where
  goal : (fm : Fin (suc m)) (fn : Fin (suc n)) →
         Vec→FinMatrix (_ ∷ Fin→VecMatrix (λ z → M (suc z))) fm fn ≡ M fm fn
  goal zero zero = refl
  goal zero (suc fn) i = FinVec→Vec→FinVec (λ z → M zero (suc z)) i fn
  goal (suc fm) fn i = FinMatrix→VecMatrix→FinMatrix (λ z → M (suc z)) i fm fn

VecMatrix→FinMatrix→VecMatrix : {A : Set ℓ} {m n : ℕ} (M : VecMatrix A m n) → Fin→VecMatrix (Vec→FinMatrix M) ≡ M
VecMatrix→FinMatrix→VecMatrix {m = zero} [] = refl
VecMatrix→FinMatrix→VecMatrix {m = suc m} (M ∷ MS) i = Vec→FinVec→Vec M i ∷ VecMatrix→FinMatrix→VecMatrix MS i

FinMatrixIsoVecMatrix : (A : Set ℓ) (m n : ℕ) → Iso (FinMatrix A m n) (VecMatrix A m n)
FinMatrixIsoVecMatrix A m n =
  iso Fin→VecMatrix Vec→FinMatrix VecMatrix→FinMatrix→VecMatrix FinMatrix→VecMatrix→FinMatrix

FinMatrix≃VecMatrix : (A : Set ℓ) (m n : ℕ) → FinMatrix A m n ≃ VecMatrix A m n
FinMatrix≃VecMatrix A m n = isoToEquiv (FinMatrixIsoVecMatrix A m n)
\end{code}

%<*eqmatrix>
\begin{code}
FinMatrix≡VecMatrix : (A : Set ℓ) (m n : ℕ) → FinMatrix A m n ≡ VecMatrix A m n
FinMatrix≡VecMatrix A m n = ua (FinMatrix≃VecMatrix A m n)
\end{code}
%</eqmatrix>

\begin{code}
import Cubical.Data.Nat hiding (_+_)

module _ (R' : CommRing {ℓ}) where

  open CommRingStr (snd R') renaming ( is-set to isSetR )

  private R = ⟨ R' ⟩
  \end{code}

  %<*addfinmatrix>
  \begin{code}
  addFinMatrix : ∀ {m n} → (M N : FinMatrix R m n) → FinMatrix R m n
  addFinMatrix M N k l = M k l + N k l

  addFinMatrixComm : ∀ {m n} → (M N : FinMatrix R m n) →
                     addFinMatrix M N ≡ addFinMatrix N M
  addFinMatrixComm M N i k l = +-comm (M k l) (N k l) i
  \end{code}
  %</addfinmatrix>

  %<*addvecmatrix>
  \begin{code}
  T : Set ℓ → Set ℓ
  T X = Σ[ _+_ ∈ (X → X → X) ] ((x y : X) → x + y ≡ y + x)

  TVecMatrix : ∀ {m n} → T (VecMatrix R m n)
  TVecMatrix {m} {n} = subst T  (FinMatrix≡VecMatrix R m n)
                                (addFinMatrix , addFinMatrixComm)

  addVecMatrix : ∀ {m n} → (M N : VecMatrix R m n) → VecMatrix R m n
  addVecMatrix = fst TVecMatrix

  addVecMatrixComm : ∀ {m n} → (M N : VecMatrix R m n) →
                     addVecMatrix M N ≡ addVecMatrix N M
  addVecMatrixComm = snd TVecMatrix
  \end{code}
  %</addvecmatrix>

  -- More direct definition of addition for VecMatrix:

  %<*addvecmatrix2>
  \begin{code}
  addVec : ∀ {m} → Vec R m → Vec R m → Vec R m
  addVec [] [] = []
  addVec (x ∷ xs) (y ∷ ys) = x + y ∷ addVec xs ys

  addVecMatrix' : ∀ {m n} → (M N : VecMatrix R m n) → VecMatrix R m n
  addVecMatrix' [] [] = []
  addVecMatrix' (M ∷ MS) (N ∷ NS) = addVec M N ∷ addVecMatrix' MS NS
  \end{code}
  %</addvecmatrix2>

  \begin{code}
  addVecLem : ∀ {m} → (M N : Vec R m)
            → FinVec→Vec (λ l → lookup l M + lookup l N) ≡ addVec M N
  addVecLem {zero} [] [] = refl
  addVecLem {suc m} (x ∷ xs) (y ∷ ys) = cong (λ zs → x + y ∷ zs) (addVecLem xs ys)

  -- The key lemma relating addVecMatrix and addVecMatrix'
  \end{code}

  %<*addvecmatrixeq>
  \begin{code}
  addVecMatrixPath : ∀ {m n} → (M N : VecMatrix R m n) →
                     addVecMatrix M N ≡ addVecMatrix' M N
  \end{code}
  %</addvecmatrixeq>
  \begin{code}
  addVecMatrixPath {zero} {n} [] [] j = transp (λ i → Vec (Vec R n) 0) j []
  addVecMatrixPath {suc m} {n} (M ∷ MS) (N ∷ NS) =
    addVecMatrix (M ∷ MS) (N ∷ NS)
      ≡⟨ transportUAop₂ (FinMatrix≃VecMatrix _ _ _) addFinMatrix (M ∷ MS) (N ∷ NS) ⟩
    FinVec→Vec (λ l → lookup l M + lookup l N) ∷ _
      ≡⟨ (λ i → addVecLem M N i ∷ Fin→VecMatrix (λ k l → lookup l (lookup k MS) + lookup l (lookup k NS))) ⟩
    addVec M N ∷ _
      ≡⟨ cong (λ X → addVec M N ∷ X) (sym (transportUAop₂ (FinMatrix≃VecMatrix _ _ _) addFinMatrix MS NS) ∙ addVecMatrixPath MS NS) ⟩
    addVec M N ∷ addVecMatrix' MS NS ∎

  -- We then directly get the properties about addVecMatrix'
  \end{code}
  %<*addvecmatrixcomm>
  \begin{code}
  addVecMatrixComm' : ∀ {m n} → (M N : VecMatrix R m n) →
                      addVecMatrix' M N ≡ addVecMatrix' N M
  \end{code}
  %</addvecmatrixcomm>
  \begin{code}
  addVecMatrixComm' M N = sym (addVecMatrixPath M N) ∙ addVecMatrixComm M N ∙ addVecMatrixPath N M

\end{code}
